/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define TYPECHECK_OPERANDS \
    kite_object_t *rhs = args->builtin_data.listvalue.car; \
    kite_object_t *exc; \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
    }
    
#define TYPECHECK_RIGHT_OPERAND(t) \
    kite_object_t *rhs = args->builtin_data.listvalue.car; \
    kite_object_t *exc; \
    if (rhs->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Second operand is not of the requested type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
    }

#define TYPECHECK_INT_OPERAND TYPECHECK_RIGHT_OPERAND(OBJ_INTEGER)
#define TYPECHECK_METHOD_OPERAND TYPECHECK_RIGHT_OPERAND(OBJ_METHOD)

#define PERFORM_LIST_BOOLEAN_OP(op, c) \
    kite_object_t *obj; \
    TYPECHECK_OPERANDS; \
    \
    if (kite_list_count(thd, this) op kite_list_count(thd, rhs)) { \
        kite_vm_return(thd, kite_new_boolean(thd, TRUE)); \
    } \
    \
    while(this && rhs) { \
        kite_vm_call_operator(thd, this->builtin_data.listvalue.car, \
            c, rhs->builtin_data.listvalue.car, TRUE); \
        obj = kite_vm_pop(thd); \
        \
        if (obj->type == OBJ_BOOLEAN) { \
            if (obj->builtin_data.intvalue == 0) { \
                kite_dereference_object(obj); \
                kite_vm_return(thd, kite_new_boolean(thd, FALSE)); \
            } \
            kite_dereference_object(obj); \
        } else { \
            exc = kite_new_exception(thd, "System.exceptions.InvalidArgument", \
                                "Comparison did not result in a boolean."); \
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        } \
        \
        this = this->builtin_data.listvalue.cdr; \
        rhs = this->builtin_data.listvalue.cdr; \
    } \
    \
    kite_vm_return(thd, kite_new_boolean(thd, TRUE));

#define CHECK_SINGLE_INDEX \
    int index = rhs->builtin_data.intvalue; \
    if (index < 0 || index >= kite_list_count(thd, this)) { \
        exc = kite_new_exception(thd, "System.exceptions.ArrayBounds", \
                            "Provided index is out of bounds."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
    }
    
KITE_CLASS_METHOD(List_add) {
    kite_object_t *cur = this, *rhs = args;
    kite_object_t *secondhalf = rhs->builtin_data.listvalue.car;
    
    if (cur->builtin_data.listvalue.car == NULL) {
        cur->builtin_data.listvalue.car = kite_reference_object(secondhalf);
        kite_vm_return(thd, this);
    }
    
    while(cur->builtin_data.listvalue.cdr != NULL) {
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    if (secondhalf->type == OBJ_LIST) {
        secondhalf = kite_reference_object(secondhalf);
    } else {
        secondhalf = kite_new_list(thd);
        secondhalf->builtin_data.listvalue.car = 
            kite_reference_object(rhs->builtin_data.listvalue.car);
    }
    
    cur->builtin_data.listvalue.cdr = secondhalf;
    kite_vm_return(thd, this);
}

KITE_CLASS_METHOD(List_lessthan) {
    PERFORM_LIST_BOOLEAN_OP(<, OP_LESS_THAN);
}

KITE_CLASS_METHOD(List_lessequals) {
    PERFORM_LIST_BOOLEAN_OP(<=, OP_LESS_OR_EQUALS);
}

KITE_CLASS_METHOD(List_equals) {
    PERFORM_LIST_BOOLEAN_OP(==, OP_EQUALS);
}

KITE_CLASS_METHOD(List_greaterequals) {
    PERFORM_LIST_BOOLEAN_OP(>=, OP_GREATER_OR_EQUALS);
}

KITE_CLASS_METHOD(List_greaterthan) {
    PERFORM_LIST_BOOLEAN_OP(>, OP_GREATER_THAN);
}

KITE_CLASS_METHOD(List_notequals) {
    PERFORM_LIST_BOOLEAN_OP(!=, OP_NOT_EQUALS);
}

KITE_CLASS_METHOD(List_deref) {
    TYPECHECK_INT_OPERAND;
    CHECK_SINGLE_INDEX;
    
    while(index > 0) {
        index--;
        this = this->builtin_data.listvalue.cdr;
    }
    
    kite_vm_return(thd, this->builtin_data.listvalue.car);
}

KITE_CLASS_METHOD(List_str) {
    KITE_NO_ARGS;
    kite_object_t *cur = this;
    kite_object_t *stringrep;
    char *ret = strdup("[");
    
    while(cur && cur->builtin_data.listvalue.car) {
        stringrep = kite_string_object(thd, cur->builtin_data.listvalue.car);
        ret = realloc(ret, strlen(ret) + 
            strlen(stringrep->builtin_data.stringvalue.string) + 3);
        strcat(ret, stringrep->builtin_data.stringvalue.string);
        kite_dereference_object(stringrep);
        if (cur->builtin_data.listvalue.cdr) {
            strcat(ret, ", ");
        }
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    strcat(ret, "]");
    stringrep = kite_new_string(thd, ret);
    free(ret);
    kite_vm_return(thd, stringrep);
}

KITE_CLASS_METHOD(List_count) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_new_integer(thd, kite_list_count(thd, this)));
}

KITE_CLASS_METHOD(List_head) {
    KITE_NO_ARGS;
    kite_object_t *ret = this->builtin_data.listvalue.car;
    if (ret == NULL) ret = kite_new_null(thd);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(List_tail) {
    KITE_NO_ARGS;
    kite_object_t *ret = this->builtin_data.listvalue.cdr;
    if (ret == NULL) ret = kite_new_list(thd);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(List_reset) {
    KITE_NO_ARGS;
    if (this->builtin_data.listvalue.iterator) {
        kite_dereference_object(this->builtin_data.listvalue.iterator);
    }
    this->builtin_data.listvalue.iterator = kite_reference_object(this);
}

KITE_CLASS_METHOD(List_cur) {
    kite_object_t *pos = this->builtin_data.listvalue.iterator;
    KITE_NO_ARGS;
    
    if (pos == NULL || pos->builtin_data.listvalue.car == NULL) {
        kite_vm_return(thd, kite_new_null(thd));
    }
    
    kite_vm_return(thd, pos->builtin_data.listvalue.car);
}

KITE_CLASS_METHOD(List_next) {
    KITE_NO_ARGS;
    
    if (this->builtin_data.listvalue.iterator == NULL || 
        this->builtin_data.listvalue.iterator->builtin_data.listvalue.car == NULL) {
        kite_vm_return(thd, kite_new_boolean(thd, FALSE));
    } else {
        kite_object_t *tmp = this->builtin_data.listvalue.iterator;
        this->builtin_data.listvalue.iterator = 
            kite_reference_object(tmp->builtin_data.listvalue.cdr);
        kite_dereference_object(tmp);
        if (this->builtin_data.listvalue.iterator == NULL || 
            this->builtin_data.listvalue.iterator->builtin_data.listvalue.car == NULL) {
            kite_vm_return(thd, kite_new_boolean(thd, FALSE));
        } else {
            (void)kite_reference_object(this);
            kite_vm_return(thd, kite_new_boolean(thd, TRUE));
        }
    }
}

KITE_CLASS_METHOD(List_map) {
    kite_object_t *ret, *cur = this, *curret, *methodargs;
    TYPECHECK_METHOD_OPERAND;
    
    ret = curret = kite_new_list(thd);
    while(cur) {
        methodargs = kite_list_object(thd, cur->builtin_data.listvalue.car);
        kite_vm_call_object(thd, this, rhs, methodargs);
        curret->builtin_data.listvalue.car = kite_vm_pop(thd);
        if (cur->builtin_data.listvalue.cdr) {
            curret->builtin_data.listvalue.cdr = kite_new_list(thd);
        }
        
        kite_dereference_object(methodargs);
        cur = cur->builtin_data.listvalue.cdr;
        curret = curret->builtin_data.listvalue.cdr;
    }
    
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(List_reduce) {
    kite_object_t *ret, *cur = this, *methodargs;
    TYPECHECK_METHOD_OPERAND;
    
    ret = kite_new_null(thd);
    while(cur) {
        methodargs = kite_new_list(thd);
        methodargs->builtin_data.listvalue.car = kite_reference_object(ret);
        methodargs->builtin_data.listvalue.cdr = 
            kite_list_object(thd, cur->builtin_data.listvalue.car);
        kite_vm_call_object(thd, this, rhs, methodargs);
        kite_dereference_object(ret);
        ret = kite_vm_pop(thd);
        
        kite_dereference_object(methodargs);
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(List_removeAt) {
    TYPECHECK_INT_OPERAND;
    CHECK_SINGLE_INDEX;
    
    kite_object_t *cur = this;
    while(index > 1) {
        index--;
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    kite_object_t *desired = cur->builtin_data.listvalue.cdr;
    if (index == 0) desired = this;
    kite_object_t *tail = desired->builtin_data.listvalue.cdr;
    if (tail) {
        kite_dereference_object(cur->builtin_data.listvalue.car);
        desired->builtin_data.listvalue.car = 
            kite_reference_object(tail->builtin_data.listvalue.car);
        desired->builtin_data.listvalue.cdr = 
            kite_reference_object(tail->builtin_data.listvalue.cdr);
        kite_dereference_object(tail);
    } else {
        if (desired != this) kite_dereference_object(desired);
        else {
            kite_dereference_object(desired->builtin_data.listvalue.car);
            desired->builtin_data.listvalue.car = NULL;
        }
        cur->builtin_data.listvalue.cdr = NULL;
    }
    kite_vm_return(thd, this);
}

KITE_CLASS_METHOD(List_prepend) {
    kite_object_t *arg = args->builtin_data.listvalue.car;
    kite_object_t *ret;
    
    if (this->builtin_data.listvalue.car) {
        ret = kite_new_list(thd);
        ret->builtin_data.listvalue.car = kite_reference_object(this->builtin_data.listvalue.car);
        ret->builtin_data.listvalue.cdr = kite_reference_object(this->builtin_data.listvalue.cdr);
        kite_dereference_object(this->builtin_data.listvalue.car);
        kite_dereference_object(this->builtin_data.listvalue.cdr);
        this->builtin_data.listvalue.car = kite_reference_object(arg);
        this->builtin_data.listvalue.cdr = ret;
    } else {
        this->builtin_data.listvalue.car = kite_reference_object(arg);
    }
    kite_vm_return(thd, this);
}

KITE_CLASS_METHOD(List_sublist) {
    TYPECHECK_INT_OPERAND;
    CHECK_SINGLE_INDEX;
    int length;
    kite_object_t *ret, *cur = this, *curret;
    
    if (args->builtin_data.listvalue.cdr) {
        args = args->builtin_data.listvalue.cdr;
        TYPECHECK_INT_OPERAND;
        length = rhs->builtin_data.intvalue;
    } else {
        length = INT_MAX;
    }
    
    /* Skip beginning */
    while(index > 0) {
        index--;
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    /* Create new list object */
    ret = kite_new_list(thd);
    curret = ret;
    while(cur && length > 0) {
        curret->builtin_data.listvalue.car = 
            kite_reference_object(cur->builtin_data.listvalue.car);
        if (cur->builtin_data.listvalue.cdr && length > 1) {
            curret->builtin_data.listvalue.cdr = kite_new_list(thd);
        }
        cur = cur->builtin_data.listvalue.cdr;
        curret = curret->builtin_data.listvalue.cdr;
        length--;
    }
    
    kite_vm_return(thd, ret);
}

KITE_MODULE_INITIALIZER {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "list");
    
    /*
     * Add operators
     */
    kite_add_operator(thread, newclass, OP_ADD, 
        kite_new_method(thread, FUNC_COMPILED, List_add, 1));
    kite_add_operator(thread, newclass, OP_LESS_THAN, 
        kite_new_method(thread, FUNC_COMPILED, List_lessthan, 1));
    kite_add_operator(thread, newclass, OP_LESS_OR_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, List_lessequals, 1));
    kite_add_operator(thread, newclass, OP_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, List_equals, 1));
    kite_add_operator(thread, newclass, OP_GREATER_OR_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, List_greaterequals, 1));
    kite_add_operator(thread, newclass, OP_GREATER_THAN, 
        kite_new_method(thread, FUNC_COMPILED, List_greaterthan, 1));
    kite_add_operator(thread, newclass, OP_NOT_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, List_notequals, 1));
    kite_add_operator(thread, newclass, OP_ARRAY_DEREF, 
        kite_new_method(thread, FUNC_COMPILED, List_deref, 1));
    kite_add_operator(thread, newclass, OP_MAP, 
        kite_new_method(thread, FUNC_COMPILED, List_map, 1));
    kite_add_operator(thread, newclass, OP_REDUCE, 
        kite_new_method(thread, FUNC_COMPILED, List_reduce, 1));
    
    /* Add methods. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method(thread, FUNC_COMPILED, List_str, 0));
    kite_add_method(thread, newclass, "reset", 
        kite_new_method(thread, FUNC_COMPILED, List_reset, 0));
    kite_add_method(thread, newclass, "cur", 
        kite_new_method(thread, FUNC_COMPILED, List_cur, 0));
    kite_add_method(thread, newclass, "next", 
        kite_new_method(thread, FUNC_COMPILED, List_next, 0));
    kite_add_method(thread, newclass, "count", 
        kite_new_method(thread, FUNC_COMPILED, List_count, 0));
    kite_add_method(thread, newclass, "head", 
        kite_new_method(thread, FUNC_COMPILED, List_head, 0));
    kite_add_method(thread, newclass, "tail", 
        kite_new_method(thread, FUNC_COMPILED, List_tail, 0));
    kite_add_method(thread, newclass, "append", 
        kite_new_method(thread, FUNC_COMPILED, List_add, 1));
    kite_add_method(thread, newclass, "getIndex", 
        kite_new_method(thread, FUNC_COMPILED, List_deref, 1));
    kite_add_method(thread, newclass, "prepend", 
        kite_new_method(thread, FUNC_COMPILED, List_prepend, 1));
    kite_add_method(thread, newclass, "sublist", 
        kite_new_method(thread, FUNC_COMPILED, List_sublist, 1));
    kite_add_method(thread, newclass, "sublist", 
        kite_new_method(thread, FUNC_COMPILED, List_sublist, 2));
    kite_add_method(thread, newclass, "removeAt", 
        kite_new_method(thread, FUNC_COMPILED, List_removeAt, 1));
        
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "list", 1);
    kite_set_property(thread, parent, "list", newclass);
}

