/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define TYPECHECK_OPERANDS \
    kite_object_t *rhs = args->builtin_data.listvalue.car; \
    kite_object_t *exc; \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
    }

#define PERFORM_STRING_BOOLEAN_OPERATION(op) \
    TYPECHECK_OPERANDS; \
    int result = strcmp(this->builtin_data.stringvalue.string, \
                        rhs->builtin_data.stringvalue.string) op 0; \
    kite_vm_return(thd, kite_new_boolean(thd, result));
    
#define TYPECHECK_INT_OPERAND \
    kite_object_t *rhs = args->builtin_data.listvalue.car; \
    kite_object_t *exc; \
    if (rhs->type != OBJ_INTEGER) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Second operand must be an integer."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
    }

#define CHECK_SINGLE_INDEX \
    int index = rhs->builtin_data.intvalue; \
    if (index < 0 || index >= this->builtin_data.stringvalue.length) { \
        exc = kite_new_exception(thd, "System.exceptions.ArrayBounds", \
                                 "Provided index is out of bounds."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
    }
        
KITE_CLASS_METHOD(String_add) {
    kite_object_t *ret;
    TYPECHECK_OPERANDS;
    char *buf = strdup(this->builtin_data.stringvalue.string);
    buf = realloc(buf, strlen(buf) + 
            strlen(rhs->builtin_data.stringvalue.string) + 1);
    strcat(buf, rhs->builtin_data.stringvalue.string);
    ret = kite_new_string(thd, buf);
    free(buf);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(String_multiply) {
    kite_object_t *ret;
    TYPECHECK_INT_OPERAND;
    int num = rhs->builtin_data.intvalue;
    if (num == 0) {
        kite_vm_return(thd, kite_new_string(thd, ""));
    } else if (num < 0) {
        exc = kite_new_exception(thd, "System.exceptions.InvalidArgument",
                            "Provided argument is invalid.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    
    char *buf = strdup(this->builtin_data.stringvalue.string);
    buf = realloc(buf, strlen(buf) * num + 1);
    for( ; num > 1; num-- ) {
        strcat(buf, this->builtin_data.stringvalue.string);
    }
    ret = kite_new_string(thd, buf);
    free(buf);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(String_lessthan) {
    PERFORM_STRING_BOOLEAN_OPERATION(<);
}

KITE_CLASS_METHOD(String_lessequals) {
    PERFORM_STRING_BOOLEAN_OPERATION(<=);
}

KITE_CLASS_METHOD(String_equals) {
    PERFORM_STRING_BOOLEAN_OPERATION(==);
}

KITE_CLASS_METHOD(String_greaterequals) {
    PERFORM_STRING_BOOLEAN_OPERATION(>=);
}

KITE_CLASS_METHOD(String_greaterthan) {
    PERFORM_STRING_BOOLEAN_OPERATION(>);
}

KITE_CLASS_METHOD(String_notequals) {
    PERFORM_STRING_BOOLEAN_OPERATION(!=);
}

KITE_CLASS_METHOD(String_deref) {
    TYPECHECK_INT_OPERAND;
    int index = rhs->builtin_data.intvalue;
    if (index < 0 || 
        index >= (signed)strlen(this->builtin_data.stringvalue.string)) {
        exc = kite_new_exception(thd, "System.exceptions.ArrayBounds",
                                      "Provided index is out of bounds.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    
    char buf[2];
    buf[0] = this->builtin_data.stringvalue.string[index];
    buf[1] = 0;
    kite_vm_return(thd, kite_new_string(thd, buf));
}

KITE_CLASS_METHOD(String_str) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_reference_object(this));
}

KITE_CLASS_METHOD(String_int) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_int_object(thd, this));
}

KITE_CLASS_METHOD(String_float) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_float_object(thd, this));
}

KITE_CLASS_METHOD(String_bool) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_boolean_object(thd, this));
}

KITE_CLASS_METHOD(String_length) {
    KITE_NO_ARGS;
    kite_vm_return(thd, 
        kite_new_integer(thd, this->builtin_data.stringvalue.length));
}

KITE_CLASS_METHOD(String_lower) {
    KITE_NO_ARGS;
    char *newstr = strdup(this->builtin_data.stringvalue.string);
    char *cur = newstr;
    kite_object_t *ret;
    
    while(*cur) *cur++ = tolower(*cur);
    ret = kite_new_string(thd, newstr);
    free(newstr);
    
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

KITE_CLASS_METHOD(String_upper) {
    KITE_NO_ARGS;
    char *newstr = strdup(this->builtin_data.stringvalue.string);
    char *cur = newstr;
    kite_object_t *ret;
    
    while(*cur) *cur++ = toupper(*cur);
    ret = kite_new_string(thd, newstr);
    free(newstr);
    
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

KITE_CLASS_METHOD(String_substring) {
    TYPECHECK_INT_OPERAND;
    CHECK_SINGLE_INDEX;
    int length;
    kite_object_t *ret;
    char *tmp;
    
    if (args->builtin_data.listvalue.cdr) {
        args = args->builtin_data.listvalue.cdr;
        TYPECHECK_INT_OPERAND;
        length = rhs->builtin_data.intvalue;
    } else {
        length = INT_MAX;
    }
    
    if (length > this->builtin_data.stringvalue.length) {
        length = this->builtin_data.stringvalue.length;
    }
    
    tmp = (char*)calloc(length + 1, 1);
    strncpy(tmp, this->builtin_data.stringvalue.string + index, length);
    ret = kite_new_string(thd, tmp);
    free(tmp);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(String_split) {
    kite_object_t *delimiter, *limit, *ret;
    int max = INT_MAX; 
    char *cur;

    KITE_GET_METHOD_ARGUMENT(delimiter, 1);
    KITE_GET_METHOD_ARGUMENT(limit, 2);

    if (!KITE_IS_TYPE(delimiter, OBJ_STRING) || 
        (limit && !KITE_IS_TYPE(limit, OBJ_INTEGER)))
    {
        kite_object_t *exc;
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                 "Operands must be of (string, integer)");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }

    if (limit) {
        max = limit->builtin_data.intvalue;
    }

    cur = this->builtin_data.stringvalue.string;
    ret = kite_new_list(thd);
    while(cur && max > 0) {
        char *tmp = strstr(cur, delimiter->builtin_data.stringvalue.string);
        char *ele;
        if (tmp && tmp != cur && max > 1) {
            ele = calloc(1, tmp - cur + 1);
            strncpy(ele, cur, tmp - cur);
            tmp++;
        } else if (tmp == cur && max > 1) {
            ele = calloc(1, 1);
            *ele = 0;
            tmp++;
        } else {
            ele = calloc(1, strlen(cur) + 1);
            strcpy(ele, cur);
        }
        kite_append_list(thd, ret, kite_new_string(thd, ele));
        free(ele);

        cur = tmp;
        max--;
    }
    
    kite_vm_return(thd, ret);
}

#define CHECK_TYPE(object, t) \
    if (object == NULL) { \
        exc = kite_new_exception(thd, "System.exceptions.InvalidArgument", \
            "Not enough objects have been passed in for the format " \
            "string given."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
    } \
    if (object->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
            "Object passed to format is of invalid type for the format " \
            "specifier provided."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
    }

#define FORMAT(out, type, val) \
    fmtstr = (char*)calloc(length + 1, 1); \
    strncpy(fmtstr, cur - length + 1, length); \
    asprintf(&out, fmtstr, \
        (type)cur_fmt->builtin_data.listvalue.car->builtin_data.val); \
    //free(fmtstr);
    
KITE_CLASS_METHOD(String_format) {
    kite_object_t *format_list = args->builtin_data.listvalue.car;
    char *output = NULL, *ret_string = NULL, *cur, *fmtstr;
    int length = 0;
    kite_object_t *cur_fmt, *exc;
    
    if (format_list->type != OBJ_LIST) {
        format_list = kite_list_object(thd, format_list);
    }
    
    cur = this->builtin_data.stringvalue.string;
    cur_fmt = format_list;
    while(*cur && cur_fmt) {
        if (*cur == '%') {
            cur++;
            length++;
cont_parse:
            while(*cur && *cur != '%' && !isalpha(*cur)) {
                cur++;
                length++;
            }
            if (*cur == '%') goto cont_parse;
            
            length++;
            switch(*cur) {
                case 'd':
                case 'i':
                case 'o':
                case 'u':
                case 'x':
                case 'X':
                    /* Integer conversion. */
                    CHECK_TYPE(cur_fmt->builtin_data.listvalue.car, OBJ_INTEGER);
                    FORMAT(output, int, intvalue);
                    break;
                case 'D':
                case 'O':
                case 'U':
                    /* Long conversion. */
                    CHECK_TYPE(cur_fmt->builtin_data.listvalue.car, OBJ_INTEGER);
                    FORMAT(output, long, intvalue);
                    break;
                case 'e':
                case 'E':
                case 'F':
                case 'f':
                case 'G':
                case 'g':
                case 'A':
                case 'a':
                    /* Double conversion. */
                    CHECK_TYPE(cur_fmt->builtin_data.listvalue.car, OBJ_FLOAT);
                    FORMAT(output, double, floatvalue);
                    break;
                case 's':
                    /* String conversion */
                    CHECK_TYPE(cur_fmt->builtin_data.listvalue.car, OBJ_STRING);
                    FORMAT(output, char*, stringvalue.string);
                    break;
                default:
                    /* Invalid format string. */
                    exc = kite_new_exception(thd, 
                                        "System.exceptions.InvalidArgument",
                                        "Provided format string is invalid.");
                    kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), 
                                        TRUE);
                    break;
            }
            
            length = 0;
            if (ret_string == NULL) {
                ret_string = output;
            } else {
                ret_string = realloc(ret_string, strlen(ret_string) + 
                                     strlen(output) + 1);
                strcat(ret_string, output);
                free(output);
            }
            cur++;
            cur_fmt = cur_fmt->builtin_data.listvalue.cdr;
            continue;
        }
        
        cur++;
        length++;
    }
    
    if (cur_fmt != NULL && cur_fmt->builtin_data.listvalue.car != NULL && !*cur) {
        exc = kite_new_exception(thd, 
                            "System.exceptions.InvalidArgument",
                            "Provided format string has fewer format "
                            "specifiers than required.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), 
                            TRUE);
    } else if (length > 0 || *cur) {
        asprintf(&output, cur - length);
        if (ret_string != NULL) {
            ret_string = 
                realloc(ret_string, strlen(ret_string) + strlen(output) + 1);
            strcat(ret_string, output);
            free(output);
        } else {
            ret_string = output;
        }
    } else if (!ret_string) {
        ret_string = strdup("");
    }
    
    cur_fmt = kite_new_string(thd, ret_string);
    free(ret_string);
    kite_vm_return(thd, cur_fmt);
}
    
KITE_MODULE_INITIALIZER {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "string");
    
    /*
     * Add operators
     */
    kite_add_operator(thread, newclass, OP_ADD, 
        kite_new_method(thread, FUNC_COMPILED, String_add, 1));
    kite_add_operator(thread, newclass, OP_MULTIPLY, 
        kite_new_method(thread, FUNC_COMPILED, String_multiply, 1));
    kite_add_operator(thread, newclass, OP_LESS_THAN, 
        kite_new_method(thread, FUNC_COMPILED, String_lessthan, 1));
    kite_add_operator(thread, newclass, OP_LESS_OR_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, String_lessequals, 1));
    kite_add_operator(thread, newclass, OP_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, String_equals, 1));
    kite_add_operator(thread, newclass, OP_GREATER_OR_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, String_greaterequals, 1));
    kite_add_operator(thread, newclass, OP_GREATER_THAN, 
        kite_new_method(thread, FUNC_COMPILED, String_greaterthan, 1));
    kite_add_operator(thread, newclass, OP_NOT_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, String_notequals, 1));
    kite_add_operator(thread, newclass, OP_ARRAY_DEREF, 
        kite_new_method(thread, FUNC_COMPILED, String_deref, 1));
    
    /* Add methods. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method(thread, FUNC_COMPILED, String_str, 0));
    kite_add_method(thread, newclass, "int", 
        kite_new_method(thread, FUNC_COMPILED, String_int, 0));
    kite_add_method(thread, newclass, "float", 
        kite_new_method(thread, FUNC_COMPILED, String_float, 0));
    kite_add_method(thread, newclass, "bool", 
        kite_new_method(thread, FUNC_COMPILED, String_bool, 0));
    kite_add_method(thread, newclass, "length", 
        kite_new_method(thread, FUNC_COMPILED, String_length, 0));
    kite_add_method(thread, newclass, "lower", 
        kite_new_method(thread, FUNC_COMPILED, String_lower, 0));
    kite_add_method(thread, newclass, "upper", 
        kite_new_method(thread, FUNC_COMPILED, String_upper, 0));
    kite_add_method(thread, newclass, "append", 
        kite_new_method(thread, FUNC_COMPILED, String_add, 1));
    kite_add_method(thread, newclass, "repeat", 
        kite_new_method(thread, FUNC_COMPILED, String_multiply, 1));
    kite_add_method(thread, newclass, "charAt", 
        kite_new_method(thread, FUNC_COMPILED, String_deref, 1));
    kite_add_method(thread, newclass, "substring", 
        kite_new_method(thread, FUNC_COMPILED, String_substring, 1));
    kite_add_method(thread, newclass, "substring", 
        kite_new_method(thread, FUNC_COMPILED, String_substring, 2));
    kite_add_method(thread, newclass, "split", 
        kite_new_method(thread, FUNC_COMPILED, String_split, 1));
    kite_add_method(thread, newclass, "split", 
        kite_new_method(thread, FUNC_COMPILED, String_split, 2));
    kite_add_method(thread, newclass, "format", 
        kite_new_method(thread, FUNC_COMPILED, String_format, 1));
        
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "string", 1);
    kite_set_property(thread, parent, "string", newclass);
}
