/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#ifdef WIN32
#include <windows.h>
#else
#include <dlfcn.h>
#endif
#include "kite_object.h"
#include "kite_gc.h"

#define NULL_OBJECT "System.null"
#define INTEGER_OBJECT "System.integer"
#define FLOAT_OBJECT "System.float"
#define STRING_OBJECT "System.string"
#define LIST_OBJECT "System.list"
#define METHOD_OBJECT "System.method"
#define BOOLEAN_OBJECT "System.boolean"

/*****************************************************************************
 * Create a new null object
 ****************************************************************************/
kite_object_t *kite_new_null(kite_thread_t *thd) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, NULL_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->type = OBJ_NULL;
    return newobj;
}

/*****************************************************************************
 * Create a new integer object
 ****************************************************************************/
kite_object_t *kite_new_integer(kite_thread_t *thd, long val) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, INTEGER_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->builtin_data.intvalue = val;
    newobj->type = OBJ_INTEGER;
    return newobj;
}

/*****************************************************************************
 * Create a new floating-point object
 ****************************************************************************/
kite_object_t *kite_new_float(kite_thread_t *thd, double val) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, FLOAT_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->builtin_data.floatvalue = val;
    newobj->type = OBJ_FLOAT;
    return newobj;
}

/*****************************************************************************
 * Create a new string object
 ****************************************************************************/
kite_object_t *kite_new_string(kite_thread_t *thd, char *str) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, STRING_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->builtin_data.stringvalue.length = strlen(str);
    newobj->builtin_data.stringvalue.string = strdup(str);
    newobj->type = OBJ_STRING;
    return newobj;
}

/*****************************************************************************
 * Create a new string object
 ****************************************************************************/
kite_object_t *kite_new_ident(kite_thread_t *thd, char *str) {
    kite_object_t *newobj = NULL;
    assert(thd != NULL);
    
    newobj = kite_new_class(thd, NULL, "");
    assert(newobj != NULL);
    
    newobj->builtin_data.stringvalue.length = strlen(str);
    newobj->builtin_data.stringvalue.string = strdup(str);
    newobj->type = OBJ_IDENT;
    return newobj;
}

/*****************************************************************************
 * Create a new list object
 ****************************************************************************/
kite_object_t *kite_new_list(kite_thread_t *thd) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, LIST_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
        
    newobj->builtin_data.listvalue.car = NULL;
    newobj->builtin_data.listvalue.cdr = NULL;
    newobj->builtin_data.listvalue.iterator = NULL;
    newobj->type = OBJ_LIST;
    return newobj;
}

/*****************************************************************************
 * Create a new boolean object
 ****************************************************************************/
kite_object_t *kite_new_boolean(kite_thread_t *thd, int val) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, BOOLEAN_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->builtin_data.intvalue = val;
    newobj->type = OBJ_BOOLEAN;
    return newobj;
}

/*****************************************************************************
 * Create a new method object
 ****************************************************************************/
kite_object_t *kite_new_method(kite_thread_t *thd, 
                                         int type, 
                                        void *ptr,
                                      int numargs) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);

    cls = kite_dereference_and_load(thd, METHOD_OBJECT);
    assert(cls != NULL);

    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
        
    newobj->builtin_data.funcvalue.functype = type;
    newobj->builtin_data.funcvalue.funcptr = ptr;
    newobj->builtin_data.funcvalue.numargs = numargs;
    newobj->builtin_data.funcvalue.this = NULL;
    newobj->type = OBJ_METHOD;
    return newobj;
}

/*****************************************************************************
 * Create a new class object
 ****************************************************************************/
kite_object_t *kite_new_class(kite_thread_t *thd, kite_object_t *inherits,
                              char *name) {
    kite_object_t *newobj = NULL;
    assert(thd != NULL);

    newobj = kite_gc_incremental(thd);
    if (!newobj) {
        newobj = (kite_object_t*)malloc(sizeof(kite_object_t));
        assert(newobj != NULL);
        memset(newobj, 0, sizeof(kite_object_t));
    }
    
    newobj->object_data.name = strdup(name);
    newobj->object_data.inherit_from = kite_reference_object(inherits);
    newobj->object_data.creatable = 1;
    pthread_mutex_init(&newobj->object_data.mtx, NULL);
    newobj->vm = thd->vm;
    newobj->type = OBJ_CLASS;
    newobj->refcount = 1;
    if (!newobj->gc_entry) {
        newobj->gc_entry = malloc(sizeof(kite_list_t));
        newobj->gc_entry->obj = newobj;
        newobj->gc_entry->prev = newobj->gc_entry->next = NULL;
    }
    kite_list_add_end(&thd->vm->gc_begin, &thd->vm->gc_end, newobj->gc_entry);

    return newobj;
}

/*****************************************************************************
 * Create a new instance of a class
 ****************************************************************************/
kite_object_t *kite_new_instance(kite_thread_t *thd, kite_object_t *class) {
    kite_object_t *newobj = NULL;
    assert(class != NULL);
    
    newobj = kite_gc_incremental(thd);
    if (!newobj) {
        newobj = (kite_object_t*)malloc(sizeof(kite_object_t));
        assert(newobj != NULL);
        memset(newobj, 0, sizeof(kite_object_t));
    }

    newobj->object_data.name = NULL;
    newobj->object_data.inherit_from = kite_reference_object(class);
    pthread_mutex_init(&newobj->object_data.mtx, NULL);
    newobj->vm = class->vm;
    newobj->type = OBJ_INSTANCE;
    newobj->refcount = 1;
    if (!newobj->gc_entry) {
        newobj->gc_entry = malloc(sizeof(kite_list_t));
        newobj->gc_entry->obj = newobj;
        newobj->gc_entry->prev = newobj->gc_entry->next = NULL;
    }
    kite_list_add_end(&thd->vm->gc_begin, &thd->vm->gc_end, newobj->gc_entry);
    
    return newobj;
}

/*****************************************************************************
 * Create a new instance of a class (and call appropriate constructor)
 ****************************************************************************/
kite_object_t *
kite_new_instance_with_constructor(kite_thread_t *thd, kite_object_t *class,
                                   kite_object_t *args) {
    kite_object_t *inst = kite_new_instance(thd, class);
    
    /* Call constructor, if one is defined. */
    kite_vm_call_constructor(thd, inst, args);
    
    return inst;
}

/*****************************************************************************
 * Create new exception object
 ****************************************************************************/
kite_object_t *kite_new_exception(kite_thread_t *thd, char *name, char *msg) {
    kite_object_t *cls = kite_dereference_and_load(thd, name);
    kite_object_t *mesg = kite_new_string(thd, msg);
    kite_object_t *args = kite_list_object(thd, mesg);
    kite_object_t *ret = kite_new_instance_with_constructor(thd, cls, args);
    
    kite_dereference_object(args);
    kite_dereference_object(mesg);
    return ret;
}

/*****************************************************************************
 * Destroy given object
 ****************************************************************************/
void
kite_destruct_object_nofree(kite_thread_t *thd, kite_object_t **obj, int r)
{
    kite_object_t *us = *obj;
    int i;
    
    /* Ignore if null. */
    if (us == NULL) {
        *obj = NULL;
        return;
    }
    assert(us->gc_entry != NULL);

    if (us->parent && r) {
        kite_dereference_object(us->parent);
        us->parent = NULL;
    }

    /* Call destructor */
    if (us->object_data.destructor && r) {
        kite_dereference_object(us->object_data.destructor);
    }

    /* Free list objects and string values */
    if (us->type == OBJ_LIST && r) {
        kite_dereference_object(us->builtin_data.listvalue.car);
        kite_dereference_object(us->builtin_data.listvalue.cdr);
        kite_dereference_object(us->builtin_data.listvalue.iterator);
    } else if (us->type == OBJ_STRING || us->type == OBJ_IDENT) {
        free(us->builtin_data.stringvalue.string);
    }

    /* Free everything else */
    if (us->object_data.name) {
        free(us->object_data.name);
    }
    if (us->object_data.properties)
        kite_destruct_symtab(thd, &us->object_data.properties, r);
    if (us->type == OBJ_CLASS || us->type == OBJ_METHOD ||
        us->type == OBJ_INSTANCE) {
        if (us->builtin_data.funcvalue.functype == FUNC_BYTECODE &&
                us->builtin_data.funcvalue.this == NULL) {
            kite_free_instruction_list(thd, 
                    (kite_opcode_t*)us->builtin_data.funcvalue.funcptr);
            us->builtin_data.funcvalue.funcptr = NULL;
        }
    }
    if (r) kite_dereference_object(us->object_data.inherit_from);
    for(i = 0; i < NUM_OPERATORS; i++) {
        if (us->object_data.operators[i] && r) {
            kite_dereference_object(us->object_data.operators[i]);
        }
    }

    if (us->object_data.dylib_hndl)
    {
#ifdef WIN32
        FreeLibrary(us->object_data.dylib_hndl);
#else
        dlclose(us->object_data.dylib_hndl);
#endif
    }
    
    kite_list_remove(&thd->vm->gc_begin, &thd->vm->gc_end, (*obj)->gc_entry);
}

/*****************************************************************************
 * Destroy given object
 ****************************************************************************/
void kite_destruct_object(kite_thread_t *thd, kite_object_t **obj, int r)
{
    kite_destruct_object_nofree(thd, obj, r);
    free((*obj)->gc_entry);
    (*obj)->gc_entry = NULL;

    free(*obj);
    *obj = NULL;
}

