/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "kite_object.h"

/*****************************************************************************
 * Find rightmost element
 ****************************************************************************/
static kite_symtab_t **kite_rightmost_symtab(kite_symtab_t **start) {
    assert(*start != NULL);
    while((*start)->right) start = &(*start)->right;
    return start;
}

/*****************************************************************************
 * Copy a symbol table.
 ****************************************************************************/
void kite_copy_symtab(kite_symtab_t **dest, kite_symtab_t *src) {
    assert(src != NULL);
    
    *dest = malloc(sizeof(kite_symtab_t));
    assert(*dest != NULL);
    
    if (src->left) kite_copy_symtab(&(*dest)->left, src->left);
    else (*dest)->left = NULL;
    if (src->right) kite_copy_symtab(&(*dest)->right, src->right);
    else (*dest)->right = NULL;
    
    (*dest)->name = strdup(src->name);
    (*dest)->global = src->global;
    (*dest)->value = kite_reference_object(src->value);
}

/*****************************************************************************
 * Destroy a symbol table.
 ****************************************************************************/
void kite_destruct_symtab(kite_thread_t *thd, kite_symtab_t **obj, int r)
{
    assert(*obj != NULL);
    
    if ((*obj)->left) kite_destruct_symtab(thd, &(*obj)->left, r);
    if ((*obj)->right) kite_destruct_symtab(thd, &(*obj)->right, r);
    
    free((*obj)->name);
    if (r) kite_dereference_object((*obj)->value);
    free(*obj);
    *obj = NULL;
}

/*****************************************************************************
 * Insert object into symbol table
 ****************************************************************************/
void 
kite_symtab_insert(kite_thread_t *thd, kite_symtab_t **st, kite_symtab_t *obj) {
    kite_symtab_t **pos = st;
    int r;
    
    /* Empty table case */
    if (!*pos) {
        *pos = obj;
        return;
    }
    
    /* Find correct location */
    while(*pos) {
        r = strcmp(obj->name, (*pos)->name);
        if (!r && (*pos)->global == obj->global) {
            /* Replace element */
            obj->left = (*pos)->left;
            obj->right = (*pos)->right;
            kite_dereference_object((*pos)->value);
            free((*pos)->name);
            free(*pos);
            *pos = obj;
            return;
        } else if (r < 0 || (!r && (*pos)->global < obj->global)) {
            pos = &(*pos)->left;
            if (!*pos) {
                kite_symtab_insert(thd, pos, obj);
                return;
            }
        } else {
            pos = &(*pos)->right;
            if (!*pos) {
                kite_symtab_insert(thd, pos, obj);
                return;
            }
        }
    }
}

/*****************************************************************************
 * Remove object from symbol table.
 ****************************************************************************/
void kite_symtab_remove(kite_thread_t *thd, kite_symtab_t **st, char *name) {
    kite_symtab_t **replacement_root, *us = *st, *new_right;
    int r;
    
    assert(*st != NULL);
    r = strcmp((*st)->name, name);
    
    if (!r) {
        replacement_root = kite_rightmost_symtab(&(*st)->left);
        kite_dereference_object(us->value);
        free(us->name);
        
        *st = *replacement_root;
        if (*st) {
            new_right = (*replacement_root)->left;
            (*st)->left = us->left;
            (*st)->right = us->right;
            *replacement_root = new_right;
        }
        
        free(us);
    } else if (r < 0) {
        kite_symtab_remove(thd, &(*st)->left, name);
    } else {
        kite_symtab_remove(thd, &(*st)->right, name);
    }
}

