/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#define KITE_EXECUTE_C
#include "kite_vm.h"
#include "kite_opcodes.h"
#include "objs/kite_object.h"

/*
 * Valid operator strings
 */
char *OPERATOR_STRINGS[] = {
    "+",
    "-",
    "*",
    "/",
    "%",
    "unary +",
    "unary -",
    "=",
    "==",
    "!=",
    "<",
    ">",
    "<=",
    ">=",
    "and",
    "or",
    "not",
    "xor",
    "<<",
    ">>",
    "map",
    "reduce",
    "[]",
    "|",
    NULL,
};

/*****************************************************************************
 * Kite opcode: Push onto execution stack
 ****************************************************************************/
static void 
kite_vm_execute_push(kite_thread_t *thd, kite_opcode_push *ins)
{
    assert(ins->obj != NULL);
    kite_vm_push(thd, ins->obj);
}

/*****************************************************************************
 * Kite opcode: Place function arguments into symbol table.
 ****************************************************************************/
static void 
kite_vm_execute_funcargs(kite_thread_t *thd, kite_opcode_funcargs *ins)
{
    kite_object_t *args, *cur_array, *cur_args, *arg;
    kite_symtab_t *symt;
    kite_funccall_info_t *stackentry;
    assert(ins->args != NULL);
    
    args = kite_vm_pop(thd);
    cur_array = args;
    cur_args = ins->args;
    
    while(cur_array && cur_args) {
        stackentry = (kite_funccall_info_t*)thd->func_stack->obj;
        arg = cur_args->builtin_data.listvalue.car;
        
        if (arg != NULL) {
            symt = malloc(sizeof(kite_symtab_t));
            symt->name = strdup(arg->builtin_data.stringvalue.string);
            symt->global = 0;
            symt->left = symt->right = NULL;
            symt->value = 
                kite_reference_object(cur_array->builtin_data.listvalue.car);
            kite_symtab_insert(thd, &stackentry->locals, symt);
        }
    
        cur_array = cur_array->builtin_data.listvalue.cdr;
        cur_args = cur_args->builtin_data.listvalue.cdr;
    }
    
    assert(cur_array == NULL && cur_args == NULL);
}

/*****************************************************************************
 * Kite opcode: Dereference variable
 ****************************************************************************/
static void 
kite_vm_execute_deref_1(kite_thread_t *thd, kite_opcode_deref_1 *ins)
{
    kite_object_t *ident, *res = NULL, *exc;
    kite_stack_t *cur = thd->func_stack;
    kite_symtab_t *found;
    kite_funccall_info_t *stackentry;
    int global = 0, searchthis = 0;
    char buf[1024];
    assert(cur != NULL);
    
    ident = kite_vm_pop(thd);
    assert(ident->type == OBJ_IDENT);
    
    if (strcmp("this", ident->builtin_data.stringvalue.string) == 0) {
        res = kite_reference_object(((kite_funccall_info_t*)cur->obj)->this);
    } else if (strcmp("__exc", ident->builtin_data.stringvalue.string) == 0) {
        res = kite_reference_object(thd->exception);
    } else if (strcmp("base", ident->builtin_data.stringvalue.string) == 0) {
        kite_object_t *this = ((kite_funccall_info_t*)cur->obj)->this;
        if (this->object_data.inherit_from) {
            res = kite_reference_object(this->object_data.inherit_from);
        } else {
            res = kite_new_null(thd);
        }
    } else {
        do {
            stackentry = (kite_funccall_info_t*)cur->obj;
            if (searchthis) {
                KITE_FIND_ANY_IN_SYMTAB(found, 
                    stackentry->this->object_data.properties,
                    ident->builtin_data.stringvalue.string);
            } else {
                KITE_FIND_ANY_IN_SYMTAB(found, stackentry->locals,
                    ident->builtin_data.stringvalue.string);
            }
            if (found) {
                if (found->value) res = found->value;
                else res = found->value = kite_new_null(thd);
                break;
            } else {
                if (searchthis) {
                    searchthis = 0;
                    cur = cur->next;
                } else {
                    searchthis = 1;
                }
            }
        } while(cur != NULL);
        
        /* If not in the stack, find in the root namespace */
        if (res == NULL) {
            KITE_FIND_PROPERTY(found, thd->vm->root_package, 
                ident->builtin_data.stringvalue.string, 1);
            if (found) {
                res = found->value;
            }
        }
        
        if (res == NULL && ins->create) {
            /* Create entry in symbol table */
            stackentry = (kite_funccall_info_t*)thd->func_stack->obj;
            found = malloc(sizeof(kite_symtab_t));
            found->name = strdup(ident->builtin_data.stringvalue.string);
            found->global = 0;
            found->left = found->right = NULL;
            found->value = kite_new_null(thd);
            kite_symtab_insert(thd, &stackentry->locals, found);
            res = found->value;
        } else if (res == NULL && !ins->create) {
            PUSH_FUNC_STACK(NULL);
            sprintf(buf, "Could not find property or variable %s", 
                ident->builtin_data.stringvalue.string);
            exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                     buf);
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            POP_FUNC_STACK;
        }
    }

    kite_vm_push(thd, res);
    //kite_dereference_object(res);
    kite_dereference_object(ident);
}

/*****************************************************************************
 * Kite opcode: Define variable
 ****************************************************************************/
static void 
kite_vm_execute_defprop(kite_thread_t *thd, kite_opcode_defprop *ins)
{
    kite_object_t *ident, *res = NULL;
    kite_symtab_t *found;
    kite_funccall_info_t *stackentry;

    (void)ins;
    ident = kite_vm_pop(thd);
    assert(ident->type == OBJ_IDENT);

    /* Create entry in symbol table */
    stackentry = (kite_funccall_info_t*)thd->func_stack->obj;
    KITE_FIND_IN_SYMTAB(found, stackentry->locals, ident->builtin_data.stringvalue.string, 0)
    if (found == NULL) {
        found = malloc(sizeof(kite_symtab_t));
        found->name = strdup(ident->builtin_data.stringvalue.string);
        found->global = 0;
        found->left = found->right = NULL;
        found->value = kite_new_null(thd);
        kite_symtab_insert(thd, &stackentry->locals, found);
    }

    kite_dereference_object(ident);
}

/*****************************************************************************
 * Kite opcode: Dereference object property
 ****************************************************************************/
static void 
kite_vm_execute_deref_2(kite_thread_t *thd, kite_opcode_deref_2 *ins)
{
    kite_object_t *ident, *obj, *res, *curr, *exc;
    kite_symtab_t *entry;
    char buf[1024];
    char *name;
    
    ident = kite_vm_pop(thd);
    obj = kite_vm_pop(thd);
    curr = obj;
    name = ident->builtin_data.stringvalue.string;
    
    do {
        KITE_FIND_ANY_IN_SYMTAB(entry, curr->object_data.properties, name);
        if (entry) {
            if ((!entry->global && obj != curr) ||
                (entry->value && entry->value->type == OBJ_METHOD && 
                 obj != curr)) {
                break;
            }
            
            if (!entry->value) entry->value = kite_new_null(thd);
            kite_vm_push(thd, entry->value);
            kite_dereference_object(ident);
            kite_dereference_object(obj);
            return;
        }
        
        curr = curr->object_data.inherit_from;
    } while(curr != NULL);
    
    if (curr) {
        kite_add_property(thd, obj, name, 0);
        
        if (entry->value && entry->value->type == OBJ_METHOD) {
            res = kite_new_method(thd, 0, NULL, 0);
            memcpy(&res->builtin_data.funcvalue,
                   &entry->value->builtin_data.funcvalue,
                   sizeof(kite_function_t));
            res->builtin_data.funcvalue.this = obj;
        } else {
            res = kite_new_null(thd);
        }
        kite_set_property(thd, obj, name, res);
        kite_vm_push(thd, res);
        kite_dereference_object(res);
        kite_dereference_object(ident);
        kite_dereference_object(obj);
        return;
    }
    
    {
        PUSH_FUNC_STACK(obj);
        sprintf(buf, "Could not find property %s", name);
        exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                 buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        POP_FUNC_STACK;
    }
}

/*****************************************************************************
 * Kite opcode: Call method
 ****************************************************************************/
static void 
kite_vm_execute_call(kite_thread_t *thd, kite_opcode_call *ins)
{
    kite_object_t *args, *method, *this, *exc, *tmp;
    kite_stack_t *curstack;

    args = kite_vm_pop(thd);
    method = kite_vm_pop(thd);
    this = kite_vm_pop(thd);

    curstack = thd->running_stack;

    if (method->type == OBJ_METHOD && method->builtin_data.funcvalue.this) {
        this = method->builtin_data.funcvalue.this;
    }
    
    PUSH_FUNC_STACK(this);
    
    if (method->type == OBJ_METHOD) {
        if (kite_list_count(thd, args) != 
            method->builtin_data.funcvalue.numargs) {
            exc = kite_new_exception(thd, "System.exceptions.InvalidArgument", 
                                     "Specified arguments do not match.");
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        }
        kite_vm_call_object(thd, this, method, args);
    } else if (method->type == OBJ_IDENT) {
        kite_vm_call_method(thd, this, method->builtin_data.stringvalue.string,
                            args, TRUE);
    } else {
        assert(0);
    }
    
    POP_FUNC_STACK;
    kite_dereference_object(args);
    kite_dereference_object(method);

    /* Remove anything other than the return value from the top of the stack */
    if (thd->running_stack != curstack) {
        kite_object_t *ret = kite_vm_pop(thd);
        while(thd->running_stack != curstack) {
            tmp = kite_vm_pop(thd);
            kite_dereference_object(tmp);
        }
        kite_vm_push(thd, ret);
        kite_dereference_object(ret);
    }
}

/*****************************************************************************
 * Kite opcode: Execute code
 ****************************************************************************/
static void 
kite_vm_execute_eval(kite_thread_t *thd, kite_opcode_eval *ins)
{
    kite_object_t *code;

    (void)ins;
    code = kite_vm_pop(thd);
    kite_vm_execute_user_method(thd, code->builtin_data.funcvalue.funcptr);
    kite_dereference_object(code);                
}

/*****************************************************************************
 * Kite opcode: Construct a list
 ****************************************************************************/
static void 
kite_vm_execute_list_cons(kite_thread_t *thd, kite_opcode_list_cons *ins)
{
    kite_object_t *car, *cdr, *res;

    (void)ins;

    cdr = kite_vm_pop(thd);
    car = kite_vm_pop(thd);
    res = kite_new_list(thd);
    
    if (cdr->type != OBJ_LIST) {
        res->builtin_data.listvalue.cdr = kite_list_object(thd, cdr);
        kite_dereference_object(cdr);
    } else {
        res->builtin_data.listvalue.cdr = cdr;
    }
    res->builtin_data.listvalue.car = car;
    
    kite_vm_push(thd, res);
    kite_dereference_object(res);
}

/*****************************************************************************
 * Kite opcode: Construct a list
 ****************************************************************************/
static void 
kite_vm_execute_list_cons_1(kite_thread_t *thd, kite_opcode_list_cons *ins)
{
    kite_object_t *car, *res;
    
    (void)ins;

    car = kite_vm_pop(thd);
    res = kite_new_list(thd);
    res->builtin_data.listvalue.car = car;
    
    kite_vm_push(thd, res);
    kite_dereference_object(res);
}

/*****************************************************************************
 * Kite opcode: Duplicate top of stack
 ****************************************************************************/
static void 
kite_vm_execute_dupe_top(kite_thread_t *thd, kite_opcode_dupe_top *ins)
{
    kite_object_t *top;
    
    (void)ins;

    top = kite_vm_pop(thd);
    kite_vm_push(thd, top);
    kite_vm_push(thd, top);
    kite_dereference_object(top);
}
    
/*****************************************************************************
 * Kite opcode: Jump if true
 ****************************************************************************/
static void *
kite_vm_execute_jumptrue(kite_thread_t *thd, kite_opcode_jump_true *ins)
{
    kite_object_t *obj;
    void *newpc;
    assert(ins->jumpto != NULL && thd->func_stack != NULL);
    
    obj = kite_boolean_object(thd, (kite_object_t*)thd->running_stack->obj);
    if (obj->builtin_data.intvalue)
    {
        newpc = ins->jumpto;
    }
    else
    {
        newpc = ins->common.next;
    }
    kite_dereference_object(obj);
    return newpc;
}

/*****************************************************************************
 * Kite opcode: Jump if false
 ****************************************************************************/
static void *
kite_vm_execute_jumpfalse(kite_thread_t *thd, kite_opcode_jump_false *ins)
{
    kite_object_t *obj;
    void *newpc;
    assert(ins->jumpto != NULL && thd->func_stack != NULL);
    
    obj = kite_boolean_object(thd, (kite_object_t*)thd->running_stack->obj);
    if (!obj->builtin_data.intvalue)
    {
        newpc = ins->jumpto;
    }
    else
    {
        newpc = ins->common.next;
    }
    kite_dereference_object(obj);
    return newpc;
}

/*****************************************************************************
 * Kite opcode: Handle exception
 ****************************************************************************/
static void *kite_vm_execute_exception_hndl(
    kite_thread_t *thd, kite_opcode_exception_hndl *ins)
{
    kite_exception_handler_t *hndl = malloc(sizeof(kite_exception_handler_t));
    kite_thread_t *exc;
        
    hndl->tos = thd->running_stack;
    hndl->func_stack = thd->func_stack;
    
    kite_push_stack(&thd->exception_handlers, hndl);
    if ((exc = (kite_thread_t*)(long)sigsetjmp(hndl->loc, 1))) {
        while(thd->running_stack && thd->running_stack != hndl->tos) {
            kite_object_t *obj = kite_vm_pop(thd);
            kite_dereference_object(obj);
        }
        while(thd->func_stack && thd->func_stack != hndl->func_stack) {
            kite_funccall_info_t *obj = kite_pop_stack(&thd->func_stack);
            free(obj);
        }
        free(hndl);
        return ins->handler;
    } else {
        kite_vm_execute_user_method(thd, ins->common.next);
        return NULL;
    }
}

/*****************************************************************************
 * Kite opcode: Handle exception
 ****************************************************************************/
static void kite_vm_execute_pop_exception_hndl(
    kite_thread_t *thd, kite_opcode_pop_exception_hndl *ins)
{
    assert(thd->exception_handlers != NULL);

    (void)ins;

    kite_exception_handler_t *loc = kite_pop_stack(&thd->exception_handlers);
    free(loc);
}

/*****************************************************************************
 * Kite opcode: Get "this"
 ****************************************************************************/
static void kite_vm_execute_this(kite_thread_t *thd, kite_opcode_this *ins)
{
    (void)ins;
    kite_vm_push(thd, ((kite_funccall_info_t*)thd->func_stack->obj)->this);
}

/*****************************************************************************
 * Kite opcode: Execute arithmetic operation
 ****************************************************************************/
static void
kite_vm_execute_arithop(kite_thread_t *thd, kite_opcode_arithop *ins)
{
    kite_object_t *lhs, *rhs;
    
    assert(ins->operation < NUM_OPERATORS);
    rhs = kite_vm_pop(thd);
    lhs = kite_vm_pop(thd);
    
    PUSH_FUNC_STACK(lhs);    
    kite_vm_call_operator(thd, lhs, ins->operation, rhs, TRUE);
    POP_FUNC_STACK;
}

/*****************************************************************************
 * Kite opcode: Execute make operation
 ****************************************************************************/
static void
kite_vm_execute_make(kite_thread_t *thd, kite_opcode_make *ins)
{
    kite_object_t *args, *obj, *newobj;
    kite_stack_t *current_stack;

    args = kite_vm_pop(thd);
    obj = kite_vm_pop(thd);
    current_stack = thd->running_stack;

    PUSH_FUNC_STACK(obj);    
    newobj = kite_new_instance_with_constructor(thd, obj, args);
    POP_FUNC_STACK;

    if (thd->running_stack != current_stack) {
        while(thd->running_stack != current_stack) {
            obj = kite_vm_pop(thd);
            kite_dereference_object(obj);
        }
    }

    kite_vm_push(thd, newobj);
}

/*****************************************************************************
 * Kite opcode: Execute is/isof operation
 ****************************************************************************/
static void
kite_vm_execute_objis(kite_thread_t *thd, kite_opcode_objis *ins)
{
    kite_object_t *class, *obj, *cur;
    int res = FALSE;
    
    class = kite_vm_pop(thd);
    obj = kite_vm_pop(thd);
    cur = obj->object_data.inherit_from;
    
    do {
        if (cur == class) res = TRUE;
        
        cur = cur->object_data.inherit_from;
    } while(cur && !res && ins->isof);
    
    kite_vm_push(thd, kite_new_boolean(thd, res));
}

/*****************************************************************************
 * Call method on object programmatically
 ****************************************************************************/
void kite_vm_call_method(kite_thread_t *thd, kite_object_t *obj, 
                         char *name, struct kite_object_t *args, int err)
{
    kite_symtab_t *found = NULL;
    kite_object_t *curr = obj, *exc;
    char buf[1024];
    int numargs = kite_list_count(thd, args);
    
    do {
        KITE_FIND_METHOD(found, curr, name, numargs);
        if (found && found->value->type == OBJ_METHOD) {
            kite_vm_call_object(thd, obj, found->value, args);
            return;
        } else if (found) {
            sprintf(buf, "%s is not callable.", name);
            exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                                     buf);
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        }
        
        curr = curr->object_data.inherit_from;
    } while(curr);
    
    /* Call function operator if method's not found */
    if (strcmp("__construct__", name) != 0) {
        curr = kite_new_list(thd);
        curr->builtin_data.listvalue.car = kite_new_string(thd, name);
        curr->builtin_data.listvalue.cdr = kite_new_list(thd);
        curr->builtin_data.listvalue.cdr->builtin_data.listvalue.car =
            kite_reference_object(args);
            
        if (kite_vm_call_operator(thd, obj, OP_METHOD_CALL, curr, 0)) {
            kite_dereference_object(curr);
            return;
        }
        
        kite_dereference_object(curr);
    }
    
    /* Method not found, raise exception */
    if (err) {
        if (strcmp("__construct__", name) == 0) {
            sprintf(buf, "Could not find constructor that takes %d argument(s)",
                    numargs);
        } else {
            sprintf(buf, "Could not find method %s that takes %d argument(s)",
                    name, numargs);
        }
        exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                 buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
}

/*****************************************************************************
 * Call method on object programmatically
 ****************************************************************************/
void 
kite_vm_call_object(kite_thread_t *thd, struct kite_object_t *this, 
                    struct kite_object_t *method, struct kite_object_t *args)
{
    kite_compiled_func_t func;
    int freestack = 0;
    kite_object_t *oldthis = NULL;
    
    /*
     * [Ticket 5] Replace with current this if we're simply calling a parent 
     * class's method.
     */
    if (this->type != OBJ_INSTANCE && thd->func_stack && 
        thd->func_stack->next) {
        kite_funccall_info_t *current = 
            (kite_funccall_info_t*)(thd->func_stack->next->obj);
        kite_object_t *cur_this = current->this;
        while(cur_this != NULL) {
            if (cur_this == this) {
                oldthis = this;
                this = current->this;
                break;
            }
            cur_this = cur_this->object_data.inherit_from;
        }
    }
    
    assert(method->type == OBJ_METHOD || method->type == OBJ_CLASS ||
           method->type == OBJ_INSTANCE);
    if (method->builtin_data.funcvalue.functype == FUNC_COMPILED) {
        func = method->builtin_data.funcvalue.funcptr;
        func(thd, this, args);
        return;
    }
    
    /* Non-compiled method */
    if (!thd->func_stack) {
        freestack = 1;
        kite_funccall_info_t *entry = calloc(sizeof(kite_funccall_info_t), 1);
        entry->this = this;
        entry->file = "(Kite loader)";
        kite_push_stack(&thd->func_stack, entry);
    } else {
        kite_funccall_info_t *entry = 
            (kite_funccall_info_t*)thd->func_stack->obj;
        entry->this = this;
    }
    
    kite_vm_push(thd, args);
    kite_vm_execute_user_method(thd, method->builtin_data.funcvalue.funcptr);
    
    if (oldthis) {
        kite_funccall_info_t *entry = 
            (kite_funccall_info_t*)thd->func_stack->obj;
        entry->this = oldthis;
    }
    
    if (freestack) {
        free(kite_pop_stack(&thd->func_stack));
    }
}

/*****************************************************************************
 * Call constructor on object programmatically
 ****************************************************************************/
void kite_vm_call_constructor(kite_thread_t *thd, struct kite_object_t *this, 
                              struct kite_object_t *args) 
{
    kite_symtab_t *symt;
    kite_object_t *curr = this->object_data.inherit_from;
    
    /* 
     * Find __construct__ (with any arguments) in the object first. If we
     * can't find any, just ignore this call. Otherwise, call constructor
     * and have the system throw an exception if the arguments don't match
     * up.
     *
     * Note [Ticket 5]: we're only looking in the current object, not the
     * parents. Any constructors in the current object are responsible for
     * calling parent constructors as needed (by using 
     * classname.__construct__).
     *
     * Note2: We also want to spit out an error if we're trying to call a
     * constructor that takes arguments that does not exist.
     */    
    KITE_FIND_ANY_IN_SYMTAB(symt, curr->object_data.properties, 
        "__construct__");
    if (symt != NULL || kite_list_count(thd, args) > 0) {
        kite_vm_call_method(thd, this, "__construct__", args, TRUE);
        return;
    }
}

/*****************************************************************************
 * Call operator method on object programmatically
 ****************************************************************************/
int kite_vm_call_operator(kite_thread_t *thd, struct kite_object_t *this, 
                          int op, struct kite_object_t *arg, int err)
{
    kite_object_t *cur = this, *exc;
    kite_object_t *args = kite_list_object(thd, arg);
    
    do {
        if (cur->object_data.operators[op]) {
            kite_vm_call_object(thd, this, cur->object_data.operators[op], 
                                args);
            kite_dereference_object(args);
            return 1;
        }
        
        cur = cur->object_data.inherit_from;
    } while(cur);
    
    /* Operator not defined, raise exception */
    if (err) {
        exc = kite_new_exception(thd, "System.exceptions.OperatorNotDefined", 
                                 OP_TO_STRING(op));
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    return 0;
}

/*****************************************************************************
 * Handle exception thrown in user code or Kite standard library/VM
 ****************************************************************************/
void kite_vm_execute_exception(kite_thread_t *thd)
{
    kite_object_t *name = kite_object_name(thd, thd->exception);
    kite_symtab_t *msg, *trace;
    kite_object_t *cur = NULL;
    
    KITE_FIND_PROPERTY(msg, thd->exception, "msg", 0);
    KITE_FIND_PROPERTY(trace, thd->exception, "trace", 0);
    
    printf("%s: %s\n", name->builtin_data.stringvalue.string,
                       msg->value->builtin_data.stringvalue.string);
    if (trace) cur = trace->value;
    while(cur != NULL) {
        printf("  %s\n", 
            cur->builtin_data.listvalue.car->builtin_data.stringvalue.string);
        cur = cur->builtin_data.listvalue.cdr;
    }               
    kite_dereference_object(name);
    exit(0);
}

/*****************************************************************************
 * Main execution loop for Kite opcodes
 ****************************************************************************/
void *kite_vm_execute(void *ptr)
{
    kite_thread_t *thd = (kite_thread_t*)ptr;
    kite_object_t *args;
    
    if (thd->running_stack) {
        args = kite_vm_pop(thd);
    } else {
        args = kite_new_null(thd);
    }
    
    kite_vm_call_object(thd, thd->start, thd->start, args);
    kite_dereference_object(args);
    return NULL;
}

/*****************************************************************************
 * Main execution loop for Kite opcodes
 ****************************************************************************/
void kite_vm_execute_user_method(kite_thread_t *thd, void *pc)
{
    kite_opcode_t *cur = (kite_opcode_t*)pc;
    while(cur)
    {    
        /* Decode, execute */
        switch(cur->opcode)
        {
            case ARITH_OP:
            {
                kite_vm_execute_arithop(thd, (kite_opcode_arithop*)cur);
                cur = cur->next;
                break;
            }
            case PUSH:
            {
                kite_vm_execute_push(thd, (kite_opcode_push*)cur);
                cur = cur->next;
                break;
            }
            case LIST_CONS:
            {
                kite_vm_execute_list_cons(thd, (kite_opcode_list_cons*)cur);
                cur = cur->next;
                break;
            }
            case LIST_CONS_1:
            {
                kite_vm_execute_list_cons_1(thd, (kite_opcode_list_cons*)cur);
                cur = cur->next;
                break;
            }
            case DEREF_1:
            {
                kite_vm_execute_deref_1(thd, (kite_opcode_deref_1*)cur);
                cur = cur->next;
                break;
            }
            case DEREF_2:
            {
                kite_vm_execute_deref_2(thd, (kite_opcode_deref_2*)cur);
                cur = cur->next;
                break;
            }
            case CALL:
            {
                kite_vm_execute_call(thd, (kite_opcode_call*)cur);
                cur = cur->next;
                break;
            }
            case JUMP_TRUE:
            {
                cur = 
                    kite_vm_execute_jumptrue(thd, (kite_opcode_jump_true*)cur);
                break;
            }
            case JUMP_FALSE:
            {
                cur = 
                    kite_vm_execute_jumpfalse(thd, (kite_opcode_jump_false*)cur);
                break;
            }
            case EXCEPTION_HNDL:
            {
                cur = kite_vm_execute_exception_hndl(
                    thd, (kite_opcode_exception_hndl*)cur);
                break;
            }
            case POP_EXCEPTION_HNDL:
            {
                kite_vm_execute_pop_exception_hndl(
                    thd, (kite_opcode_pop_exception_hndl*)cur);
                cur = cur->next;
                break;
            }
            case DUPE_TOP:
            {
                kite_vm_execute_dupe_top(thd, (kite_opcode_dupe_top*)cur);
                cur = cur->next;
                break;
            }
            case FUNCARGS:
            {
                kite_vm_execute_funcargs(thd, (kite_opcode_funcargs*)cur);
                cur = cur->next;
                break;
            }
            case MAKE_OBJ:
            {
                kite_vm_execute_make(thd, (kite_opcode_make*)cur);
                cur = cur->next;
                break;
            }
            case EVAL_OBJ:
            {
                kite_vm_execute_eval(thd, (kite_opcode_eval*)cur);
                cur = cur->next;
                break;
            }
            case IS_ISOF:
            {
                kite_vm_execute_objis(thd, (kite_opcode_objis*)cur);
                cur = cur->next;
                break;
            }
            case DEFPROP:
            {
                kite_vm_execute_defprop(thd, (kite_opcode_defprop*)cur);
                cur = cur->next;
                break;
            }
            case THIS:
            {
                kite_vm_execute_this(thd, (kite_opcode_this*)cur);
                cur = cur->next;
                break;
            }
            case NOP:
            {
                /* Ignore. */
                cur = cur->next;
                break;
            }
            default:
            {
                /* Invalid opcode */
                assert(0);
                break;
            }
        }

    }
}
